; Copyright (C) 2023  Val Lorentz
; 
; This program is free software: you can redistribute it and/or modify it under the
; terms of the GNU Affero General Public License version 3, as published by the Free
; Software Foundation.
; 
; This program is distributed in the hope that it will be useful, but WITHOUT ANY
; WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
; PARTICULAR PURPOSE.  See the GNU Affero General Public License for more details.
; 
; You should have received a copy of the GNU Affero General Public License along with
; this program.  If not, see <http://www.gnu.org/licenses/>.

(ns jable.client
  (:require [clojure.tools.logging :refer :all]
            [clojure.java.io :as io]
            [clojure.string :as str]
            [clojure.instant :as instant]
            [jepsen [cli :as cli]
                    [client :as client]
                    [control :as c]
                    [db :as db]
                    [tests :as tests]]
            [jepsen.control.util :as cu]
            [jepsen.os.debian :as debian])
  (:import java.net.Socket))

(defn escape-tag-value [v]
  (str/replace v #"[; \\\r\n]" (fn [c]
                                 (case c
                                   ";" "\\:"
                                   " " "\\s"
                                   "\\" "\\\\"
                                   "\r" "\\r"
                                   "\n" "\\n"))))

(defn write-command [writer {tags :tags, :or {}
                             cmd :cmd
                             params :params}]
  (if (and tags (not (empty? tags)))
    (do
      (.write writer "@")
      (doseq [[i [k v]] (map-indexed vector tags)]
        (do
          (if (= i 0)
            nil
            (.write writer ";"))
          (.write writer k)
          (.write writer "=")
          (.write writer (escape-tag-value (or v "")))))
      (.write writer " "))
    nil)
  (.write writer cmd)
  (.write writer
    (if (empty? params)
      ""
      (let [first-params (pop params)
            trailing-param (last params)]
        (apply str (map #(str " " %) (conj first-params (str ":" trailing-param)))))))
  (.write writer "\r\n"))

(defn send-command [client command]
  (info "send:" command)
  (write-command (:writer client) command)
  (.flush (:writer client)))

(defn send-commands [client commands]
  (doseq [command commands]
    (info "send:" command)
    (write-command (:writer client) command))
  (.flush (:writer client)))

(defn unescape-tag-value [v]
  (str/replace v #"\\(.)" (fn [[_ c]]
                            (case c
                              ":" ";"
                              "s" " "
                              "\\" "\\"
                              "r" "\r"
                              "n" "\n"
                              c))))

(defn parse-tag [[k v]]
  (if v
    [k (unescape-tag-value v)]
    [k ""]))

(defn parse-tags [s]
  (as-> s v
    (str/split v #";")
    (map #(str/split % #"=" 2) v)
    (map parse-tag v)
    (into {} v)))

(def line-re
  #"^(?:@([^ ]+) )?(?::([^ ]+) )?([^ ]+)(?: (.*?))??(?: :(.*))?$")

(defn parse-command [line]
  (if line
    (let [[_ tags source cmd params trailing] (re-matches line-re line)]
      {:tags (if tags (parse-tags tags) {})
       :source source
       :cmd cmd
       :params (let [params (if (empty? params)
                              []
                              (str/split params #" "))]
                 (if trailing
                   (conj params trailing)
                   params))})
    nil))

(defn read-response-batch [client label ref-tags acc]
  (let [cmd (parse-command (.readLine (:reader client)))
        [inner-ref-tag & outer-ref-tags] ref-tags]
    (assert cmd "got empty line / end of stream within a batch")
    (info "recv:" cmd)
    (if (= (:cmd cmd) "BATCH")
      (let [[ref-tag batch-type & _] (:params cmd)]
        (case (first ref-tag)
          ; new (inner) batch, add its ref-tag to the stack
          \+ (read-response-batch client label
                                  (conj ref-tags (str/replace-first ref-tag "+" ""))
                                  (conj acc cmd))
          ; closing batch
          \- (if (= [(str "-" inner-ref-tag)] (:params cmd))
               (if (empty? outer-ref-tags)
                 ; end of outer-most batch, return commands
                 (conj acc cmd)
                 ; end of inner-most batch, continue
                 (read-response-batch client label outer-ref-tags (conj acc cmd)))
               ; end of unrelated batch, ignore
               (read-response-batch client label ref-tags acc))))
      ; note: this assumes outer batches don't get new messages while an inner batch is open.
      (if (= (get (:tags cmd) "batch" nil) inner-ref-tag)
        ; add this command, continue
        (read-response-batch client label ref-tags (conj acc cmd))
        ; command not in batch, ignore and continue
        (read-response-batch client label ref-tags acc)))))


(defn read-response [client label]
  (let [cmd (parse-command (.readLine (:reader client)))]
    (assert cmd "got empty line / end of stream")
    (info "recv:" cmd)
    (if (= (get (:tags cmd) "label" nil) label)
      (if (= (:cmd cmd) "BATCH")
        (let [[ref-tag & _] (:params cmd)]
          ; read & return the batch
          (read-response-batch client label (conj '() (str/replace-first ref-tag "+" "")) [cmd]))
        [cmd]) ; return only this line
      (read-response client label)))) ; ignore this line, skip to next


(defrecord Client [socket reader writer]
  client/Client
  (open! [this test node]
    (let [socket (Socket. node 6667)]
      (let [this (assoc this
                       :socket socket
                       :reader (io/reader socket)
                       :writer (io/writer socket)
                       :nick (str "test-" node "-" (.getLocalPort socket)))]
        (send-commands this [{:cmd "CAP" :params ["LS" "302"]}
                             {:cmd "CAP" :params ["REQ" "labeled-response batch draft/channel-rename server-time"]}
                             {:cmd "NICK" :params [(:nick this)]}
                             {:cmd "USER" :params [(str "test-" node) "*" "*" (str "test " node " port " (.getLocalPort socket))]}
                             {:tags {"label" "cap-end"} :cmd "CAP" :params ["END"]}])
        (read-response this "cap-end")  ; block until end of registration
        (info "registered")
        (send-command this {:tags {"label" "join-#chan"} :cmd "JOIN" :params ["#chan"]})
        (read-response this "join-#chan")  ; block until joined
        (info "joined")
        (Thread/sleep 1000)
        (send-command this {:tags {"label" "names"}:cmd "NAMES" :params ["#chan"]})
        (send-command this {:tags {"label" "mode-topic"} :cmd "MODE" :params ["#chan" "-t"]})
        (read-response this "mode-topic")  ; block until mode is set
        (info "set mode -t")
        (send-command this {:tags {"label" "get-mode" } :cmd "MODE" :params ["#chan"]})
        (read-response this "get-mode")  ; block
        (info "read mode")
        this)))

  (setup! [this test])

  (invoke! [this test op]
    (case (:f op)
        :read (do
                (send-command this {:tags {"label" "read-topic"}
                                    :cmd "TOPIC"
                                    :params ["#chan"]})
                (let [[cmd1 & cmds] (read-response this "read-topic")]
                  (case (:cmd cmd1)
                    ; RPL_NOTOPIC
                    "331" (assoc op :type :ok, :value nil)
                    "BATCH" (let [[topic-cmd topicwhotime-cmd batch-close] cmds]
                              ; RPL_TOPIC
                              (assert (= (:cmd topic-cmd) "332") (str "expected RPL_TOPIC: " topic-cmd))
                              ; RPL_TOPICWHOTIME
                              (assert (= (:cmd topicwhotime-cmd) "333") (str "expected RPL_TOPICWHOTIME: " topicwhotime-cmd))
                              (assert (= (:cmd batch-close) "BATCH") "topic batch close cmd")
                              (let [[my-nick chan topic] (:params topic-cmd)
                                    [my-nick chan who time] (:params topicwhotime-cmd)]
                                (let [[nick userhost] (str/split who #"!" 2)
                                      time (Integer/parseInt time)]
                                  (assoc op :type :ok, :value {:topic topic
                                                               :who nick
                                                               :time time})))))))
        :write (do
                 (send-command this {:tags {"label" "write-topic"}
                                     :cmd "TOPIC"
                                     :params ["#chan" (:value op)]})
                 (let [[cmd] (read-response this "write-topic")]
                   (case (:cmd cmd)
                    ; Sable returns ACK when two clients set the same topic value
                    ; at the same time
                    "ACK" (assoc op :type :fail, :error :noop)
                    "TOPIC" (let [[chan topic] (:params cmd)
                                  time (/ (.getTime (instant/read-instant-timestamp (get (:tags cmd) "time"))) 1000)]
                              (assoc op :type :ok, :value {:topic topic
                                                           :who (:nick this)
                                                           :time time })))))))

  (teardown! [this test])

  (close! [this test]
    (send-command this {:cmd "QUIT" :params ["closing client"]})
    (.close (:writer this))
    (.close (:reader this))
    (.close (:socket this))))
