(ns jable.client-test
  (:require [clojure.test :refer :all]
            [jable.client :refer :all]))

(deftest write-command-test
  (testing "formatting command with no param"
    (let [buf (java.io.StringWriter.)]
      (write-command buf {:cmd "AWAY" :params []})
      (is (= (.toString buf) "AWAY\r\n"))))

  (testing "formatting command with one param"
    (let [buf (java.io.StringWriter.)]
      (write-command buf {:cmd "NICK" :params ["bar"]})
      (is (= (.toString buf) "NICK :bar\r\n"))))

  (testing "formatting command with two params"
    (let [buf (java.io.StringWriter.)]
      (write-command buf {:cmd "PRIVMSG" :params ["#chan" "hello"]})
      (is (= (.toString buf) "PRIVMSG #chan :hello\r\n"))))

  (testing "formatting command with three params"
    (let [buf (java.io.StringWriter.)]
      (write-command buf {:cmd "KICK" :params ["#chan" "badperson" "bye bye"]})
      (is (= (.toString buf) "KICK #chan badperson :bye bye\r\n")))

  (testing "formatting command with empty set of tags"
    (let [buf (java.io.StringWriter.)]
      (write-command buf {:tags {} :cmd "AWAY" :params []})
      (is (= (.toString buf) "AWAY\r\n"))))

  (testing "formatting command with tag with no value"
    (let [buf (java.io.StringWriter.)]
      (write-command buf {:tags {"foo" nil} :cmd "AWAY" :params []})
      (is (= (.toString buf) "@foo= AWAY\r\n"))))

  (testing "formatting command with tag with empty value"
    (let [buf (java.io.StringWriter.)]
      (write-command buf {:tags {"foo" ""} :cmd "AWAY" :params []})
      (is (= (.toString buf) "@foo= AWAY\r\n"))))

  (testing "formatting command with tag"
    (let [buf (java.io.StringWriter.)]
      (write-command buf {:tags {"foo" "bar"} :cmd "AWAY" :params []})
      (is (= (.toString buf) "@foo=bar AWAY\r\n"))))

  (testing "formatting command with tags"
    (let [buf (java.io.StringWriter.)]
      (write-command buf {:tags {"foo" "bar" "baz" "qux"} :cmd "AWAY" :params []})
      (is (= (.toString buf) "@foo=bar;baz=qux AWAY\r\n"))))

  (testing "formatting command with tag value containing chars to escape"
    (let [buf (java.io.StringWriter.)]
      (write-command buf {:tags {"foo" "ab;cd ef\rgh\nij"} :cmd "AWAY" :params []})
      (is (= (.toString buf) "@foo=ab\\:cd\\sef\\rgh\\nij AWAY\r\n"))))))

(deftest parse-command-test
  (testing "parsing a command with no param"
    (is (= (parse-command "AWAY")
           {:tags {} :source nil :cmd "AWAY" :params []})))

  (testing "parsing a command with one param"
    (is (= (parse-command "AWAY reason")
           {:tags {} :source nil :cmd "AWAY" :params ["reason"]})))

  (testing "parsing a command with one trailing param"
    (is (= (parse-command "AWAY :the reason")
           {:tags {} :source nil :cmd "AWAY" :params ["the reason"]})))

  (testing "parsing a command with two params"
    (is (= (parse-command "PRIVMSG #chan :hi")
           {:tags {} :source nil :cmd "PRIVMSG" :params ["#chan" "hi"]})))

  (testing "parsing a command with one normal param and one trailing param"
    (is (= (parse-command "PRIVMSG #chan :hi there")
           {:tags {} :source nil :cmd "PRIVMSG" :params ["#chan" "hi there"]})))

  (testing "parsing a command with source"
    (is (= (parse-command ":nick!user@host AWAY")
           {:tags {} :source "nick!user@host" :cmd "AWAY" :params []})))

  (testing "parsing a command with a tag"
    (is (= (parse-command "@label=abc ACK")
           {:tags {"label" "abc"} :source nil :cmd "ACK" :params []})))

  (testing "parsing a command with a tag containing escaped chars"
    (is (= (parse-command "@label=ab\\:cd\\sef\\rgh\\nij\\klm ACK")
           {:tags {"label" "ab;cd ef\rgh\nijklm"} :source nil :cmd "ACK" :params []})))

  (testing "parsing a command with tags"
    (is (= (parse-command "@label=abc;time=123 ACK")
           {:tags {"label" "abc", "time", "123"} :source nil :cmd "ACK" :params []})))

  (testing "parsing a command with a tag with an empty value"
    (is (= (parse-command "@tag= ACK")
           {:tags {"tag" ""} :source nil :cmd "ACK" :params []})))

  (testing "parsing a command with a tag with no value"
    (is (= (parse-command "@tag ACK")
           {:tags {"tag" ""} :source nil :cmd "ACK" :params []})))

  (testing "parsing a command with source and tag"
    (is (= (parse-command "@label=abc :nick!user@host AWAY")
           {:tags {"label" "abc"} :source "nick!user@host" :cmd "AWAY" :params []})))

  (testing "parsing a command with all fields"
    (is (= (parse-command "@label=abc;time=123 :server. KICK #chan badperson :reason")
           {:tags {"label" "abc", "time", "123"}  :source "server." :cmd "KICK" :params ["#chan" "badperson" "reason"]}))))

(deftest read-response-test
  (testing "reading single-line response"
    (let [buf (java.io.BufferedReader. (java.io.StringReader. "@label=abc PONG\r\n"))]
      (is (= (read-response {:reader buf} "abc")
             [{:tags {"label" "abc"} :source nil :cmd "PONG" :params []}]))))

  (testing "ignoring unlabelled line"
    (let [buf (java.io.BufferedReader. (java.io.StringReader. "PONG\r\n@label=abc ACK\r\n"))]
      (is (= (read-response {:reader buf} "abc")
             [{:tags {"label" "abc"} :source nil :cmd "ACK" :params []}]))))

  (testing "ignoring line with different label"
    (let [buf (java.io.BufferedReader. (java.io.StringReader. "@label=other PONG\r\n@label=abc ACK\r\n"))]
      (is (= (read-response {:reader buf} "abc")
             [{:tags {"label" "abc"} :source nil :cmd "ACK" :params []}]))))

  (testing "reading single-command response batch"
    (let [buf (java.io.BufferedReader. (java.io.StringReader. "@label=abc BATCH +def labeled-response\r\n@batch=def PONG\r\nBATCH -def\r\n"))]
      (is (= (read-response {:reader buf} "abc")
             [{:tags {"label" "abc"} :source nil :cmd "BATCH" :params ["+def" "labeled-response"]}
              {:tags {"batch" "def"} :source nil :cmd "PONG" :params []}
              {:tags {} :source nil :cmd "BATCH" :params ["-def"]}]))))

  (testing "reading multi-command response batch"
    (let [buf (java.io.BufferedReader. (java.io.StringReader. "@label=abc BATCH +def labeled-response\r\n@batch=def PRIVMSG nick :msg 1\r\n@batch=def PRIVMSG nick :msg 2\r\nBATCH -def\r\n"))]
      (is (= (read-response {:reader buf} "abc")
             [{:tags {"label" "abc"} :source nil :cmd "BATCH" :params ["+def" "labeled-response"]}
              {:tags {"batch" "def"} :source nil :cmd "PRIVMSG" :params ["nick" "msg 1"]}
              {:tags {"batch" "def"} :source nil :cmd "PRIVMSG" :params ["nick" "msg 2"]}
              {:tags {} :source nil :cmd "BATCH" :params ["-def"]}]))))

  (testing "reading nested response batch"
    (let [buf (java.io.BufferedReader. (java.io.StringReader. "@label=abc BATCH +def labeled-response\r\n@batch=def BATCH +ghi draft/multiline\r\n@batch=ghi PRIVMSG nick :msg 1\r\n@batch=ghi PRIVMSG nick :msg 2\r\n@batch=def BATCH -ghi\r\nBATCH -def\r\n"))]
      (is (= (read-response {:reader buf} "abc")
             [{:tags {"label" "abc"} :source nil :cmd "BATCH" :params ["+def" "labeled-response"]}
              {:tags {"batch" "def"} :source nil :cmd "BATCH" :params ["+ghi" "draft/multiline"]}
              {:tags {"batch" "ghi"} :source nil :cmd "PRIVMSG" :params ["nick" "msg 1"]}
              {:tags {"batch" "ghi"} :source nil :cmd "PRIVMSG" :params ["nick" "msg 2"]}
              {:tags {"batch" "def"} :source nil :cmd "BATCH" :params ["-ghi"]}
              {:tags {} :source nil :cmd "BATCH" :params ["-def"]}]))))

  )
